"""
Initialize the app
"""

# Django
from django.utils.translation import gettext_lazy as _

__version__ = "2.7.2"
__title__ = _("Fleet Finder")
__verbose_name__ = "Fleet Finder for Alliance Auth"

__package_name__ = "aa-fleetfinder"
__app_name__ = "fleetfinder"
__app_name_verbose__ = "AA Fleet Finder"
__app_name_useragent__ = "AA-Fleet-Finder"

__github_url__ = f"https://github.com/ppfeufer/{__package_name__}"



"""
Tests for the fleetfinder.tasks module.
"""

# Standard Library
from unittest import TestCase
from unittest.mock import Mock, patch

# AA Fleet Finder
from fleetfinder.tasks import (
    _get_fleet_aggregate,
    check_fleet_adverts,
)


class TestGetFleetAggregate(TestCase):
    """
    Tests for the _get_fleet_aggregate function.
    """

    def test_returns_correct_counts_for_valid_fleet_infos(self):
        """
        Test that _get_fleet_aggregate returns correct counts for valid fleet_infos.

        :return:
        :rtype:
        """

        fleet_infos = [
            {"ship_type_name": "Cruiser"},
            {"ship_type_name": "Cruiser"},
            {"ship_type_name": "Battleship"},
        ]

        result = _get_fleet_aggregate(fleet_infos)

        assert result == {"Cruiser": 2, "Battleship": 1}

    def test_returns_empty_dict_for_empty_fleet_infos(self):
        """
        Test that _get_fleet_aggregate returns an empty dictionary for empty fleet_infos.

        :return:
        :rtype:
        """

        fleet_infos = []

        result = _get_fleet_aggregate(fleet_infos)

        assert result == {}

    def test_returns_only_valid_ship_type_names(self):
        """
        Test that _get_fleet_aggregate returns only valid ship type names.

        :return:
        :rtype:
        """

        fleet_infos = [
            {"ship_type_name": "Cruiser"},
            {},
            {"ship_type_name": None},
            {"ship_type_name": "Battleship"},
            {"other_key": "Frigate"},
        ]

        result = _get_fleet_aggregate(fleet_infos)

        assert result == {"Cruiser": 1, "Battleship": 1}


class TestCheckFleetAdvert(TestCase):
    """
    Tests for the check_fleet_adverts function.
    """

    @patch("fleetfinder.models.Fleet.objects.all")
    @patch("fleetfinder.tasks.fetch_esi_status")
    def test_processes_registered_fleets_when_available(
        self, mock_fetch_esi_status, mock_fleet_objects
    ):
        """
        Test that check_fleet_adverts processes registered fleets when ESI is available.

        :param mock_fetch_esi_status:
        :type mock_fetch_esi_status:
        :param mock_fleet_objects:
        :type mock_fleet_objects:
        :return:
        :rtype:
        """

        mock_fetch_esi_status.return_value.is_ok = True
        mock_fleet_objects.return_value.exists.return_value = True
        mock_fleet_objects.return_value.count.return_value = 2
        mock_fleet_objects.return_value.__iter__.return_value = iter([Mock(), Mock()])

        check_fleet_adverts()

        mock_fleet_objects.return_value.__iter__.assert_called_once()
        mock_fetch_esi_status.assert_called_once()

    @patch("fleetfinder.models.Fleet.objects.all")
    @patch("fleetfinder.tasks.fetch_esi_status")
    def test_logs_no_registered_fleets_when_none_exist(
        self, mock_fetch_esi_status, mock_fleet_objects
    ):
        """
        Test that check_fleet_adverts logs a message when no registered fleets exist.

        :param mock_fetch_esi_status:
        :type mock_fetch_esi_status:
        :param mock_fleet_objects:
        :type mock_fleet_objects:
        :return:
        :rtype:
        """

        mock_fetch_esi_status.return_value.is_ok = True
        mock_fleet_objects.return_value.exists.return_value = False

        check_fleet_adverts()

        mock_fleet_objects.return_value.exists.assert_called_once()
        mock_fetch_esi_status.assert_not_called()

    @patch("fleetfinder.models.Fleet.objects.all")
    @patch("fleetfinder.tasks.fetch_esi_status")
    def test_aborts_processing_when_esi_is_unavailable(
        self, mock_fetch_esi_status, mock_fleet_objects
    ):
        mock_fetch_esi_status.return_value.is_ok = False
        mock_fleet_objects.return_value.exists.return_value = True

        check_fleet_adverts()

        mock_fetch_esi_status.assert_called_once()
        mock_fleet_objects.return_value.__iter__.assert_not_called()



"""
Tasks
"""

# Standard Library
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta

# Third Party
from bravado.exception import HTTPNotFound
from celery import shared_task

# Django
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.esi import fetch_esi_status
from app_utils.logging import LoggerAddTag

# AA Fleet Finder
from fleetfinder import __title__
from fleetfinder.models import Fleet
from fleetfinder.providers import esi

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_MAX_RETRIES = 3
ESI_MAX_ERROR_COUNT = 3
ESI_ERROR_GRACE_TIME = 75

TASK_TIME_LIMIT = 120  # Stop after 2 minutes

# Params for all tasks
TASK_DEFAULT_KWARGS = {"time_limit": TASK_TIME_LIMIT, "max_retries": ESI_MAX_RETRIES}


class FleetViewAggregate:  # pylint: disable=too-few-public-methods
    """
    Helper class
    """

    def __init__(self, fleet, aggregate):
        self.fleet = fleet
        self.aggregate = aggregate


@shared_task
def _send_invitation(character_id, fleet_commander_token, fleet_id):
    """
    Open the fleet invite window in the eve client

    :param character_id:
    :param fleet_commander_token:
    :param fleet_id:
    """

    invitation = {"character_id": character_id, "role": "squad_member"}

    esi.client.Fleets.post_fleets_fleet_id_members(
        fleet_id=fleet_id,
        token=fleet_commander_token.valid_access_token(),
        invitation=invitation,
    ).result()


def _close_esi_fleet(fleet: Fleet, reason: str) -> None:
    """
    Closing registered fleet

    :param fleet:
    :param reason:
    """

    logger.info(
        msg=(
            f'Fleet "{fleet.name}" of {fleet.fleet_commander} (ESI ID: {fleet.fleet_id}) » '
            f"Closing: {reason}"
        )
    )

    fleet.delete()


def _esi_fleet_error_handling(fleet: Fleet, error_key: str) -> None:
    """
    ESI error handling

    :param fleet:
    :type fleet:
    :param error_key:
    :type error_key:
    :return:
    :rtype:
    """

    time_now = timezone.now()

    # Close ESI fleet if the consecutive error count is too high
    if (
        fleet.last_esi_error == error_key
        and fleet.last_esi_error_time
        >= (time_now - timedelta(seconds=ESI_ERROR_GRACE_TIME))
        and fleet.esi_error_count >= ESI_MAX_ERROR_COUNT
    ):
        _close_esi_fleet(fleet=fleet, reason=error_key.label)

        return

    error_count = (
        fleet.esi_error_count + 1
        if fleet.last_esi_error == error_key
        and fleet.last_esi_error_time
        >= (time_now - timedelta(seconds=ESI_ERROR_GRACE_TIME))
        else 1
    )

    logger.info(
        f'Fleet "{fleet.name}" of {fleet.fleet_commander} (ESI ID: {fleet.fleet_id}) » '
        f'Error: "{error_key.label}" ({error_count} of {ESI_MAX_ERROR_COUNT}).'
    )

    fleet.esi_error_count = error_count
    fleet.last_esi_error = error_key
    fleet.last_esi_error_time = time_now
    fleet.save()


@shared_task
def _get_fleet_aggregate(fleet_infos):
    """
    Getting numbers for fleet composition

    :param fleet_infos:
    :return:
    """

    counts = {}

    for member in fleet_infos:
        type_ = member.get("ship_type_name")

        if type_ and isinstance(type_, str) and type_.strip():
            type_ = type_.strip()  # Normalize ship type name

            if type_ in counts:
                counts[type_] += 1
            else:
                counts[type_] = 1

    return counts


def _check_for_esi_fleet(fleet: Fleet):
    """
    Check for required ESI scopes

    :param fleet:
    :type fleet:
    :return:
    :rtype:
    """

    required_scopes = ["esi-fleets.read_fleet.v1"]

    # Check if there is a fleet
    try:
        fleet_commander_id = fleet.fleet_commander.character_id
        esi_token = Token.get_token(fleet_commander_id, required_scopes)

        fleet_from_esi = esi.client.Fleets.get_characters_character_id_fleet(
            character_id=fleet_commander_id,
            token=esi_token.valid_access_token(),
        ).result()

        return {"fleet": fleet_from_esi, "token": esi_token}
    except HTTPNotFound:
        _esi_fleet_error_handling(error_key=Fleet.EsiError.NOT_IN_FLEET, fleet=fleet)
    except Exception:  # pylint: disable=broad-exception-caught
        _esi_fleet_error_handling(error_key=Fleet.EsiError.NO_FLEET, fleet=fleet)

    return False


def _process_fleet(fleet: Fleet) -> None:
    """
    Processing a fleet

    :param fleet: Fleet object to process
    :type fleet: Fleet
    :return: None
    :rtype: None
    """

    logger.info(
        f'Processing information for fleet "{fleet.name}" '
        f"of {fleet.fleet_commander} (ESI ID: {fleet.fleet_id})"
    )

    # Check if there is a fleet
    esi_fleet = _check_for_esi_fleet(fleet=fleet)

    if not esi_fleet:
        return

    # Fleet IDs don't match, FC changed fleets
    if fleet.fleet_id != esi_fleet["fleet"]["fleet_id"]:
        _esi_fleet_error_handling(
            fleet=fleet, error_key=Fleet.EsiError.FC_CHANGED_FLEET
        )
        return

    # Check if we deal with the fleet boss here
    try:
        _ = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=fleet.fleet_id,
            token=esi_fleet["token"].valid_access_token(),
        ).result()
    except Exception:  # pylint: disable=broad-exception-caught
        _esi_fleet_error_handling(fleet=fleet, error_key=Fleet.EsiError.NOT_FLEETBOSS)


@shared_task
def send_fleet_invitation(fleet_id: int, character_ids: list) -> None:
    """
    Send fleet invitations to characters through ESI
    This task sends fleet invitations to a list of character IDs using the ESI API.

    :param fleet_id: The ID of the fleet to which invitations are sent
    :type fleet_id: int
    :param character_ids: List of character IDs to invite to the fleet
    :type character_ids: list[int]
    :return: None
    :rtype: None
    """

    required_scopes = ["esi-fleets.write_fleet.v1"]
    fleet = Fleet.objects.get(fleet_id=fleet_id)
    fleet_commander_token = Token.get_token(
        character_id=fleet.fleet_commander.character_id, scopes=required_scopes
    )

    with ThreadPoolExecutor(max_workers=50) as ex:
        futures = [
            ex.submit(
                _send_invitation,
                character_id=character_id,
                fleet_commander_token=fleet_commander_token,
                fleet_id=fleet_id,
            )
            for character_id in character_ids
        ]

        for future in as_completed(futures):
            future.result()  # This will raise any exceptions that occurred


@shared_task(**{**TASK_DEFAULT_KWARGS}, **{"base": QueueOnce})
def check_fleet_adverts() -> None:
    """
    Check all registered fleets and process them

    :return: None
    :rtype: None
    """

    fleets = Fleet.objects.all()

    if not fleets.exists():
        logger.info("No registered fleets found. Nothing to do...")

        return

    logger.info(f"Processing {fleets.count()} registered fleets...")

    # Abort if ESI seems to be offline or above the error limit
    if not fetch_esi_status().is_ok:
        logger.warning("ESI doesn't seem to be available at this time. Aborting.")

        return

    for fleet in fleets:
        _process_fleet(fleet=fleet)


@shared_task
def get_fleet_composition(  # pylint: disable=too-many-locals
    fleet_id: int,
) -> FleetViewAggregate | None:
    """
    Get the composition of a fleet by its ID
    This task retrieves the composition of a fleet using its ESI ID.

    :param fleet_id: The ESI ID of the fleet to retrieve
    :type fleet_id:  int
    :return: FleetViewAggregate containing fleet members and aggregate data
    :rtype: FleetViewAggregate | None
    """

    try:
        fleet = Fleet.objects.get(fleet_id=fleet_id)
    except Fleet.DoesNotExist as exc:
        logger.error(f"Fleet with ID {fleet_id} not found")

        raise Fleet.DoesNotExist(f"Fleet with ID {fleet_id} not found.") from exc

    logger.info(
        f'Getting fleet composition for fleet "{fleet.name}" '
        f"of {fleet.fleet_commander.character_name} (ESI ID: {fleet_id})"
    )

    required_scopes = ["esi-fleets.read_fleet.v1"]

    try:
        token = Token.get_token(
            character_id=fleet.fleet_commander.character_id, scopes=required_scopes
        )

        fleet_infos = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=fleet_id, token=token.valid_access_token()
        ).result()

        # Get all unique IDs and fetch names in one call
        all_ids = {
            item_id
            for member in fleet_infos
            for item_id in [
                member["character_id"],
                member["solar_system_id"],
                member["ship_type_id"],
            ]
        }

        logger.debug(
            f"Found {len(all_ids)} unique IDs to fetch names for in fleet {fleet_id}"
        )

        # Process IDs in chunks of 1000 to avoid ESI limits.
        # ESI has a limit of 1000 IDs per request, so we will chunk the requests,
        # even though there is a theoretical limit of 768 unique IDs per fleet,
        # so we never should hit the ESI limit.
        # But to be on the safe side, we will chunk the requests in case CCP decides
        # to change the fleet limit in the future, we will use a chunk size of 1000,
        # which is the maximum allowed by ESI for the `post_universe_names` endpoint.
        chunk_size = 1000
        ids_to_name = []
        all_ids_list = list(all_ids)

        for i in range(0, len(all_ids_list), chunk_size):
            chunk = all_ids_list[i : i + chunk_size]
            chunk_result = esi.client.Universe.post_universe_names(ids=chunk).result()

            ids_to_name.extend(chunk_result)

        # Create a lookup dictionary for names
        name_lookup = {item["id"]: item["name"] for item in ids_to_name}

        # Add additional information to each fleet member
        for member in fleet_infos:
            is_fleet_boss = member["character_id"] == fleet.fleet_commander.character_id

            member.update(
                {
                    "character_name": name_lookup[member["character_id"]],
                    "solar_system_name": name_lookup[member["solar_system_id"]],
                    "ship_type_name": name_lookup[member["ship_type_id"]],
                    "is_fleet_boss": is_fleet_boss,
                }
            )

        aggregate = _get_fleet_aggregate(fleet_infos=fleet_infos)

        return FleetViewAggregate(fleet=fleet_infos, aggregate=aggregate)

    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f"Failed to get fleet composition for fleet {fleet_id}: {e}")

        raise RuntimeError(
            f"Failed to get fleet composition for fleet {fleet_id} : {str(e)}"
        ) from e



# Generated by Django 4.0.10 on 2023-09-22 18:31

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("groupmanagement", "0019_adding_restricted_to_groups"),
        ("fleetfinder", "0002_esi_error_handling_and_verbose_names"),
    ]

    operations = [
        migrations.AlterField(
            model_name="fleet",
            name="fleet_commander",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="fleetfinder_fleet_commander",
                to="eveonline.evecharacter",
                verbose_name="Fleet commander",
            ),
        ),
        migrations.AlterField(
            model_name="fleet",
            name="groups",
            field=models.ManyToManyField(
                help_text="Only selected groups will have access to the fleet. If no groups are selected the fleet will be available to all who have access to this module.",
                related_name="fleetfinder_restricted_groups",
                to="groupmanagement.authgroup",
                verbose_name="Group restrictions",
            ),
        ),
        migrations.AlterField(
            model_name="fleet",
            name="name",
            field=models.CharField(
                default="", max_length=50, verbose_name="Fleet name"
            ),
        ),
    ]



"""
Test user agent header for ESI requests
"""

# Django
from django.conf import settings
from django.test import TestCase

# Alliance Auth
from esi import __url__ as esi_url
from esi import __version__ as esi_version

# AA Fleet Finder
from fleetfinder import __app_name_useragent__, __github_url__, __version__
from fleetfinder.providers import esi


class TestUserAgent(TestCase):
    """
    Test the user agent header for ESI requests
    """

    def test_user_agent_header(self):
        """
        Test that the user agent header is set correctly for ESI requests.

        :return:
        :rtype:
        """

        operation = esi.client.Universe.get_universe_factions()

        self.assertEqual(
            operation.future.request.headers["User-Agent"],
            (
                f"{__app_name_useragent__}/{__version__} "
                f"({settings.ESI_USER_CONTACT_EMAIL}; +{__github_url__}) "
                f"Django-ESI/{esi_version} (+{esi_url})"
            ),
        )



"""
Views
"""

# Standard Library
import json

# Third Party
from bravado.exception import HTTPNotFound

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import Promise
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import character_portrait_url
from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_all_characters_from_user
from allianceauth.groupmanagement.models import AuthGroup
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Fleet Finder
from fleetfinder import __title__
from fleetfinder.models import Fleet
from fleetfinder.providers import esi
from fleetfinder.tasks import get_fleet_composition, send_fleet_invitation

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required()
@permission_required(perm="fleetfinder.access_fleetfinder")
def _get_and_validate_fleet(token: Token, character_id: int) -> tuple:
    """
    Get fleet information and validate fleet commander permissions

    :param token: Token object containing the access token
    :type token: Token
    :param character_id: The character ID of the fleet commander
    :type character_id: int
    :return: Tuple containing the fleet result, fleet commander, and fleet ID
    :rtype: tuple
    """

    fleet_result = esi.client.Fleets.get_characters_character_id_fleet(
        character_id=token.character_id, token=token.valid_access_token()
    ).result()

    logger.debug(f"Fleet result: {fleet_result}")

    fleet_commander = EveCharacter.objects.get(character_id=token.character_id)
    fleet_id = fleet_result.get("fleet_id")
    fleet_boss_id = fleet_result.get("fleet_boss_id")

    if not fleet_id:
        raise ValueError(f"No fleet found for {fleet_commander.character_name}")

    if fleet_boss_id != character_id:
        raise ValueError(f"{fleet_commander.character_name} is not the fleet boss")

    return fleet_result


@login_required()
@permission_required(perm="fleetfinder.access_fleetfinder")
def dashboard(request):
    """
    Dashboard view

    :param request:
    :return:
    """

    context = {}

    logger.info(msg=f"Module called by {request.user}")

    return render(
        request=request,
        template_name="fleetfinder/dashboard.html",
        context=context,
    )


@login_required()
@permission_required(perm="fleetfinder.access_fleetfinder")
def ajax_dashboard(request) -> JsonResponse:  # pylint: disable=too-many-locals
    """
    Ajax :: Dashboard information

    :param request:
    :return:
    """

    def _create_button_style_link(
        url: str, fa_icon_class: str, btn_title: str | Promise, btn_modifier_class: str
    ) -> str:
        """
        Helper function to create a button HTML string
        This function generates an HTML anchor tag styled as a button with an icon.

        :param url: The URL the button should link to
        :type url: str
        :param fa_icon_class: The Font Awesome class for the icon to be displayed
        :type fa_icon_class: str
        :param btn_title: The title attribute for the button, typically a translation string
        :type btn_title: str | Promise
        :param btn_modifier_class: The Bootstrap modifier class for the button styling
        :type btn_modifier_class: str
        :return: An HTML string representing the button
        :rtype: str
        """

        return (
            f'<a href="{url}" class="btn btn-sm {btn_modifier_class} ms-1" '
            f'data-bs-tooltip="aa-fleetfinder" title="{btn_title}">'
            f'<i class="{fa_icon_class}"></i></a>'
        )

    def _get_fleet_commander_information(fleet: Fleet) -> tuple[str, str]:
        """
        Helper function to get the fleet commander's HTML representation
        This function retrieves the fleet commander's name and portrait URL,
        and returns an HTML string with the portrait image and name.

        :param fleet: The Fleet object containing the fleet commander's information
        :type fleet: Fleet
        :return: A tuple containing the HTML string for the fleet commander and the name for sorting
        :rtype: tuple[str, str]
        """

        commander_name = fleet.fleet_commander.character_name
        portrait_url = character_portrait_url(
            character_id=fleet.fleet_commander.character_id, size=32
        )
        portrait_img = (
            '<img class="rounded eve-character-portrait" '
            f'src="{portrait_url}" alt="{commander_name}" loading="lazy">'
        )

        return portrait_img + commander_name, commander_name

    data = []
    groups = request.user.groups.all()
    user_characters = get_all_characters_from_user(user=request.user)
    fleets = (
        Fleet.objects.filter(
            Q(groups__group__in=groups)
            | Q(groups__isnull=True)
            | Q(fleet_commander__in=user_characters)
        )
        .distinct()
        .order_by("name")
    )

    can_manage_fleets = request.user.has_perm("fleetfinder.manage_fleets")

    for fleet in fleets:
        fleet_commander_html, fleet_commander_name = _get_fleet_commander_information(
            fleet
        )

        # Create buttons
        buttons = [
            _create_button_style_link(
                reverse("fleetfinder:join_fleet", args=[fleet.fleet_id]),
                "fa-solid fa-right-to-bracket",
                _("Join fleet"),
                "btn-success",
            )
        ]

        if can_manage_fleets:
            buttons.extend(
                [
                    _create_button_style_link(
                        reverse("fleetfinder:fleet_details", args=[fleet.fleet_id]),
                        "fa-solid fa-eye",
                        _("View fleet details"),
                        "btn-info",
                    ),
                    _create_button_style_link(
                        reverse("fleetfinder:edit_fleet", args=[fleet.fleet_id]),
                        "fa-solid fa-pen-to-square",
                        _("Edit fleet advert"),
                        "btn-warning",
                    ),
                ]
            )

        data.append(
            {
                "fleet_commander": {
                    "html": fleet_commander_html,
                    "sort": fleet_commander_name,
                },
                "fleet_name": fleet.name,
                "created_at": fleet.created_at,
                "actions": "".join(buttons),
            }
        )

    return JsonResponse(data=data, safe=False)


@login_required()
@permission_required(perm="fleetfinder.manage_fleets")
@token_required(scopes=("esi-fleets.read_fleet.v1", "esi-fleets.write_fleet.v1"))
def create_fleet(request, token):
    """
    Create fleet view

    :param request:
    :param token:
    :return:
    """

    # Validate the token and check if the character is in a fleet and is the fleet boss
    try:
        _get_and_validate_fleet(token, token.character_id)
    except (HTTPNotFound, ValueError) as ex:
        logger.debug(f"Error during fleet creation: {ex}", exc_info=True)

        if isinstance(ex, HTTPNotFound):
            error_detail = ex.swagger_result["error"]
        else:
            error_detail = str(ex)

        error_message = _(
            f"<h4>Error!</h4><p>There was an error creating the fleet: {error_detail}</p>"
        )

        messages.error(request, mark_safe(error_message))

        return redirect("fleetfinder:dashboard")

    if request.method != "POST":
        return redirect("fleetfinder:dashboard")

    auth_groups = AuthGroup.objects.filter(internal=False)
    context = {"character_id": token.character_id, "auth_groups": auth_groups}

    return render(
        request=request,
        template_name="fleetfinder/create-fleet.html",
        context=context,
    )


@login_required()
@permission_required(perm="fleetfinder.manage_fleets")
def edit_fleet(request, fleet_id):
    """
    Fleet edit view

    :param request:
    :param fleet_id:
    :return:
    """

    try:
        fleet = Fleet.objects.get(fleet_id=fleet_id)
    except Fleet.DoesNotExist:
        logger.debug(f"Fleet with ID {fleet_id} does not exist.")

        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>Fleet does not exist or is no longer available.</p>"
                )
            ),
        )

        return redirect("fleetfinder:dashboard")

    auth_groups = AuthGroup.objects.filter(internal=False)

    context = {
        "character_id": fleet.fleet_commander.character_id,
        "auth_groups": auth_groups,
        "fleet": fleet,
    }

    logger.debug("Context for fleet edit: %s", context)

    logger.info(msg=f"Fleet {fleet_id} edit view by {request.user}")

    return render(
        request=request,
        template_name="fleetfinder/edit-fleet.html",
        context=context,
    )


@login_required()
@permission_required(perm="fleetfinder.access_fleetfinder")
def join_fleet(request, fleet_id):
    """
    Join fleet view

    :param request:
    :param fleet_id:
    :return:
    """

    context = {}
    groups = request.user.groups.all()
    fleet = Fleet.objects.filter(
        Q(groups__group__in=groups) | Q(groups=None), fleet_id=fleet_id
    ).count()

    if fleet == 0:
        return redirect(to="fleetfinder:dashboard")

    if request.method == "POST":
        character_ids = request.POST.getlist(key="character_ids", default=[])
        send_fleet_invitation.delay(character_ids=character_ids, fleet_id=fleet_id)

        return redirect(to="fleetfinder:dashboard")

    characters = (
        EveCharacter.objects.filter(character_ownership__user=request.user)
        .select_related()
        .order_by("character_name")
    )

    context["characters"] = characters

    return render(
        request=request,
        template_name="fleetfinder/join-fleet.html",
        context=context,
    )


@login_required()
@permission_required("fleetfinder.manage_fleets")
def save_fleet(request):
    """
    Save fleet

    :param request:
    :return:
    """

    def _edit_or_create_fleet(
        character_id: int,
        free_move: bool,
        name: str,
        groups: list,
        motd: str = None,  # pylint: disable=unused-argument
    ) -> None:
        """
        Edit or create a fleet from a fleet in EVE Online

        :param character_id: The character ID of the fleet commander
        :type character_id: int
        :param free_move: Whether the fleet is free move or not
        :type free_move: bool
        :param name: Name of the fleet
        :type name: str
        :param groups: Groups that are allowed to access the fleet
        :type groups: list[AuthGroup]
        :param motd: Message of the Day for the fleet
        :type motd: str
        :return: None
        :rtype: None
        """

        required_scopes = ["esi-fleets.read_fleet.v1", "esi-fleets.write_fleet.v1"]
        token = Token.get_token(character_id=character_id, scopes=required_scopes)

        fleet_result = _get_and_validate_fleet(token, character_id)
        fleet_commander = EveCharacter.objects.get(character_id=character_id)
        fleet_id = fleet_result.get("fleet_id")

        fleet, created = Fleet.objects.get_or_create(
            fleet_id=fleet_id,
            defaults={
                "created_at": timezone.now(),
                # "motd": motd,
                "is_free_move": free_move,
                "fleet_commander": fleet_commander,
                "name": name,
            },
        )

        if not created:
            fleet.is_free_move = free_move
            fleet.name = name
            fleet.save()

        fleet.groups.set(groups)

        esi.client.Fleets.put_fleets_fleet_id(
            fleet_id=fleet_id,
            token=token.valid_access_token(),
            # new_settings={"is_free_move": free_move, "motd": motd},
            new_settings={"is_free_move": free_move},
        ).result()

    if request.method != "POST":
        return redirect("fleetfinder:dashboard")

    # Extract form data
    form_data = {
        "character_id": int(request.POST["character_id"]),
        "free_move": request.POST.get("free_move") == "on",
        # "motd": request.POST.get("motd", ""),
        "name": request.POST.get("name", ""),
        "groups": request.POST.getlist("groups", []),
    }

    logger.debug(f"Form data for fleet creation: {form_data}")

    try:
        _edit_or_create_fleet(**form_data)
    except HTTPNotFound as ex:
        logger.debug(f"ESI returned 404 for fleet creation: {ex}", exc_info=True)

        esi_error = ex.swagger_result.get("error", "Unknown error")
        error_message = _(
            f"<h4>Error!</h4><p>ESI returned the following error: {esi_error}</p>"
        )

        messages.error(request, mark_safe(error_message))
    except ValueError as ex:
        logger.debug(f"Value error during fleet creation: {ex}", exc_info=True)

        error_message = _(
            f"<h4>Error!</h4><p>There was an error creating the fleet: {ex}</p>"
        )

        messages.error(request, mark_safe(error_message))

    return redirect("fleetfinder:dashboard")


@login_required()
@permission_required(perm="fleetfinder.manage_fleets")
def fleet_details(request, fleet_id):
    """
    Fleet details view

    :param request:
    :param fleet_id:
    :return:
    """

    try:
        fleet = Fleet.objects.get(fleet_id=fleet_id)
    except Fleet.DoesNotExist:
        logger.debug(f"Fleet with ID {fleet_id} does not exist.")

        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>Fleet does not exist or is no longer available.</p>"
                )
            ),
        )

        return redirect("fleetfinder:dashboard")

    context = {"fleet": fleet}

    logger.info(msg=f"Fleet {fleet.fleet_id} details view called by {request.user}")

    return render(
        request=request,
        template_name="fleetfinder/fleet-details.html",
        context=context,
    )


@login_required()
@permission_required(perm="fleetfinder.manage_fleets")
def ajax_fleet_details(
    request, fleet_id  # pylint: disable=unused-argument
) -> JsonResponse:
    """
    Ajax :: Fleet Details

    :param request:
    :param fleet_id:
    """

    try:
        fleet = get_fleet_composition(fleet_id)
    except Fleet.DoesNotExist:
        logger.debug(f"Fleet with ID {fleet_id} does not exist.")

        return JsonResponse(
            data={"error": _(f"Fleet with ID {fleet_id} does not exist.")}, safe=False
        )
    except RuntimeError as ex:
        logger.debug(f"Error retrieving fleet composition: {ex}", exc_info=True)

        return JsonResponse(
            data={"error": _(f"Error retrieving fleet composition: {ex}")}, safe=False
        )

    data = {
        "fleet_member": list(fleet.fleet),
        "fleet_composition": [
            {"ship_type_name": ship, "number": number}
            for ship, number in fleet.aggregate.items()
        ],
    }

    return JsonResponse(data=data, safe=False)


@login_required()
@permission_required(perm="fleetfinder.manage_fleets")
def ajax_fleet_kick_member(request: WSGIRequest, fleet_id: int) -> JsonResponse:
    """
    Ajax :: Kick member from fleet

    :param request: WSGIRequest object containing the request data
    :type request: WSGIRequest
    :param fleet_id: The ID of the fleet from which to kick a member
    :type fleet_id: int
    :return: JsonResponse indicating success or failure of the operation
    :rtype: JsonResponse
    """

    if request.method != "POST":
        return JsonResponse(
            data={"success": False, "error": _("Method not allowed")}, status=405
        )

    try:
        fleet = Fleet.objects.get(fleet_id=fleet_id)
        data = json.loads(request.body)
        member_id = data.get("memberId")

        if not member_id:
            return JsonResponse(
                data={"success": False, "error": _("Member ID required")}, status=400
            )

        logger.debug(f"Request data for kicking member: {data}")

        token = Token.get_token(
            character_id=fleet.fleet_commander.character_id,
            scopes=["esi-fleets.write_fleet.v1"],
        )

        esi.client.Fleets.delete_fleets_fleet_id_members_member_id(
            fleet_id=fleet_id,
            member_id=member_id,
            token=token.valid_access_token(),
        ).result()

        return JsonResponse(data={"success": True}, status=200)
    except Fleet.DoesNotExist:
        return JsonResponse(
            data={"success": False, "error": _("Fleet not found")}, status=404
        )
    except (json.JSONDecodeError, ValueError):
        return JsonResponse(
            data={"success": False, "error": _("Invalid request data")}, status=400
        )
    except HTTPNotFound:
        return JsonResponse(
            data={"success": False, "error": _("Member not found in fleet")}, status=404
        )



"""
Test the views for the Fleet Finder application.
"""

# Standard Library
import json
from http import HTTPStatus
from types import SimpleNamespace
from unittest.mock import Mock, patch

# Django
from django.contrib.auth.models import Group
from django.test import TestCase
from django.urls import reverse
from django.utils.datetime_safe import datetime
from django.utils.timezone import now

# Alliance Auth
from allianceauth.groupmanagement.models import AuthGroup

# Alliance Auth (External Libs)
from app_utils.testing import create_fake_user

# AA Fleet Finder
from fleetfinder.models import Fleet


def dt_to_iso(dt: datetime) -> str:
    """
    Helper :: Convert a datetime object to ISO 8601 format.

    @see https://github.com/django/django/blob/main/django/core/serializers/json.py#L92-L98

    :param dt:
    :type dt:
    :return:
    :rtype:
    """

    r = dt.isoformat()

    if dt.microsecond:
        r = r[:23] + r[26:]

    if r.endswith("+00:00"):
        r = r.removesuffix("+00:00") + "Z"

    return r


class FleetfinderTestViews(TestCase):
    """
    Base test case for Fleet Finder views.
    This class sets up the necessary users and fleet ID for testing.
    It includes a user with the `fleetfinder.manage_fleets` permission
    and a user with `fleetfinder.access_fleetfinder` access permissions.
    The fleet ID is set to a predefined value for testing purposes.
    """

    @classmethod
    def setUp(cls):
        """
        Set up the test case.

        :return:
        :rtype:
        """

        cls.user_with_manage_perms = create_fake_user(
            character_id=1000,
            character_name="Jean Luc Picard",
            permissions=["fleetfinder.access_fleetfinder", "fleetfinder.manage_fleets"],
        )
        cls.user_with_basic_acces_perms = create_fake_user(
            character_id=1001,
            character_name="William Riker",
            permissions=["fleetfinder.access_fleetfinder"],
        )

        cls.fleet_created_at = now()

        cls.fleet = Fleet(
            fleet_id=12345,
            name="Starfleet",
            fleet_commander=cls.user_with_manage_perms.profile.main_character,
            created_at=cls.fleet_created_at,
            is_free_move=False,
        )
        cls.fleet.save()

        cls.fleet_id = 12345


class TestAjaxDashboardView(FleetfinderTestViews):
    """
    Test the ajax_dashboard view in the Fleet Finder application.
    This view is responsible for rendering the dashboard with fleet data.
    It should return a JSON response containing fleet information,
    including fleet names, commanders, and group associations.
    If no fleets are available, it should return an empty list.
    It should also filter fleets based on the user's groups.
    """

    @patch("fleetfinder.views.get_all_characters_from_user")
    def test_renders_dashboard_with_fleet_data_with_basic_access(
        self, mock_get_characters
    ):
        """
        Test that the ajax_dashboard view renders the dashboard with fleet data
        when the user has basic access permissions.

        :param mock_get_characters:
        :type mock_get_characters:
        :return:
        :rtype:
        """

        mock_get_characters.return_value = [
            self.user_with_manage_perms.profile.main_character
        ]

        fleet = self.fleet
        fleet.groups.set([])

        self.client.force_login(self.user_with_basic_acces_perms)
        url = reverse("fleetfinder:ajax_dashboard")
        join_url = reverse("fleetfinder:join_fleet", args=[self.fleet_id])
        response = self.client.get(url)

        expected_response = [
            {
                "fleet_commander": {
                    "html": '<img class="rounded eve-character-portrait" src="https://images.evetech.net/characters/1000/portrait?size=32" alt="Jean Luc Picard" loading="lazy">Jean Luc Picard',
                    "sort": "Jean Luc Picard",
                },
                "fleet_name": "Starfleet",
                "created_at": dt_to_iso(self.fleet_created_at),
                "actions": f'<a href="{join_url}" class="btn btn-sm btn-success ms-1" data-bs-tooltip="aa-fleetfinder" title="Join fleet"><i class="fa-solid fa-right-to-bracket"></i></a>',
            }
        ]

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIn("Starfleet", response.json()[0]["fleet_name"])
        self.assertIn("Jean Luc Picard", response.json()[0]["fleet_commander"]["html"])
        self.assertEqual(response.json(), expected_response)

    @patch("fleetfinder.views.get_all_characters_from_user")
    def test_renders_dashboard_with_fleet_data_with_manage_access(
        self, mock_get_characters
    ):
        """
        Test that the ajax_dashboard view renders the dashboard with fleet data
        when the user has manage access permissions.

        :param mock_get_characters:
        :type mock_get_characters:
        :return:
        :rtype:
        """

        mock_get_characters.return_value = [
            self.user_with_manage_perms.profile.main_character
        ]

        fleet = self.fleet
        fleet.groups.set([])

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:ajax_dashboard")
        join_url = reverse("fleetfinder:join_fleet", args=[self.fleet_id])
        details_url = reverse("fleetfinder:fleet_details", args=[self.fleet_id])
        edit_url = reverse("fleetfinder:edit_fleet", args=[self.fleet_id])
        response = self.client.get(url)

        expected_response = [
            {
                "fleet_commander": {
                    "html": '<img class="rounded eve-character-portrait" src="https://images.evetech.net/characters/1000/portrait?size=32" alt="Jean Luc Picard" loading="lazy">Jean Luc Picard',
                    "sort": "Jean Luc Picard",
                },
                "fleet_name": "Starfleet",
                "created_at": dt_to_iso(self.fleet_created_at),
                "actions": (
                    f'<a href="{join_url}" class="btn btn-sm btn-success ms-1" data-bs-tooltip="aa-fleetfinder" title="Join fleet"><i class="fa-solid fa-right-to-bracket"></i></a>'
                    f'<a href="{details_url}" class="btn btn-sm btn-info ms-1" data-bs-tooltip="aa-fleetfinder" title="View fleet details"><i class="fa-solid fa-eye"></i></a>'
                    f'<a href="{edit_url}" class="btn btn-sm btn-warning ms-1" data-bs-tooltip="aa-fleetfinder" title="Edit fleet advert"><i class="fa-solid fa-pen-to-square"></i></a>'
                ),
            }
        ]

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIn("Starfleet", response.json()[0]["fleet_name"])
        self.assertIn("Jean Luc Picard", response.json()[0]["fleet_commander"]["html"])
        self.assertEqual(response.json(), expected_response)

    @patch("fleetfinder.views.get_all_characters_from_user")
    def test_returns_empty_data_when_no_fleets_available(self, mock_get_characters):
        """
        Test that the ajax_dashboard view returns an empty list when no fleets are available.

        :param mock_get_characters:
        :type mock_get_characters:
        :return:
        :rtype:
        """

        mock_get_characters.return_value = [
            self.user_with_manage_perms.profile.main_character
        ]

        Fleet.objects.all().delete()  # Remove all fleets

        self.client.force_login(self.user_with_basic_acces_perms)
        url = reverse("fleetfinder:ajax_dashboard")
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(response.json(), [])

    @patch("fleetfinder.views.get_all_characters_from_user")
    def test_filters_fleets_by_user_groups(self, mock_get_characters):
        """
        Test that the ajax_dashboard view filters fleets based on the user's groups.

        :param mock_get_characters:
        :type mock_get_characters:
        :return:
        :rtype:
        """

        mock_get_characters.return_value = [
            self.user_with_manage_perms.profile.main_character
        ]

        group_obj = Group.objects.create(name="Starfleet Officers")
        auth_group, _ = AuthGroup.objects.get_or_create(group=group_obj)
        fleet = self.fleet
        fleet.groups.set([auth_group])

        self.client.force_login(self.user_with_basic_acces_perms)
        self.user_with_basic_acces_perms.groups.add(group_obj)
        url = reverse("fleetfinder:ajax_dashboard")
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIn("Starfleet", response.json()[0]["fleet_name"])


class TestFleetEditView(FleetfinderTestViews):
    """
    Test the edit_fleet view in the Fleet Finder application.
    This view is responsible for editing fleet details.
    """

    @patch("fleetfinder.views.Fleet.objects.get")
    @patch("fleetfinder.views.AuthGroup.objects.filter")
    def test_renders_edit_fleet_template_with_correct_context(
        self, mock_filter_groups, mock_get_fleet
    ):
        """
        Test that the edit_fleet view renders the correct template and context.

        :param mock_filter_groups:
        :type mock_filter_groups:
        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.return_value = self.fleet
        group1 = Mock(spec=AuthGroup)
        group1.name = "Group1"
        group2 = Mock(spec=AuthGroup)
        group2.name = "Group2"
        mock_filter_groups.return_value = [group1, group2]

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:edit_fleet", args=[self.fleet_id])
        response = self.client.get(url)

        self.assertTemplateUsed(response, "fleetfinder/edit-fleet.html")
        self.assertEqual(response.context["fleet"].name, "Starfleet")
        self.assertEqual(response.context["character_id"], 1000)
        self.assertEqual(len(response.context["auth_groups"]), 2)

    @patch("fleetfinder.views.Fleet.objects.get")
    def test_redirects_to_dashboard_if_fleet_does_not_exist(self, mock_get_fleet):
        """
        Test that the edit_fleet view redirects to the dashboard if the fleet does not exist.

        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.side_effect = Fleet.DoesNotExist

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:edit_fleet", args=[99999])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("fleetfinder:dashboard"))


class TestJoinFleetView(FleetfinderTestViews):
    """
    Test the join_fleet view in the Fleet Finder application.
    This view is responsible for allowing users to join a fleet.
    It should redirect to the fleet details page after joining.
    If the fleet does not exist, it should return a 404 status code.
    """

    @patch("fleetfinder.views.Fleet.objects.get")
    def test_join_fleet_redirects_to_dashboard(self, mock_get_fleet):
        """
        Test that the join_fleet view redirects to the dashboard after joining a fleet.

        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.return_value = Fleet(fleet_id=self.fleet_id)

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:join_fleet", args=[self.fleet_id])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "fleetfinder/join-fleet.html")

    @patch("fleetfinder.views.Fleet.objects.get")
    def test_join_fleet_handles_non_existent_fleet(self, mock_get_fleet):
        """
        Test that the join_fleet view handles a non-existent fleet correctly.

        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.side_effect = Fleet.DoesNotExist

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:join_fleet", args=[123456])  # Non-existent fleet ID
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, reverse("fleetfinder:dashboard"))


class TestFleetDetailsView(FleetfinderTestViews):
    """
    Test the fleet_details view in the Fleet Finder application.
    This view is responsible for rendering the fleet details page.
    It should render the correct template and require the user to have
    the 'fleetfinder.manage_fleets' permission to access it.
    If the fleet does not exist, it should return a 404 status code.
    """

    @patch("fleetfinder.views.Fleet.objects.get")
    def test_fleet_details_renders_correct_template(self, mock_get_fleet):
        """
        Test that the fleet_details view renders the correct template.

        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.return_value = Fleet(fleet_id=self.fleet_id)

        self.client.force_login(self.user_with_manage_perms)
        url = reverse("fleetfinder:fleet_details", args=[self.fleet_id])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "fleetfinder/fleet-details.html")

    @patch("fleetfinder.views.Fleet.objects.get")
    def test_fleet_details_requires_manage_permission(self, mock_get_fleet):
        """
        Test that the fleet_details view requires the user to have the 'fleetfinder.manage_fleets' permission.

        :param mock_get_fleet:
        :type mock_get_fleet:
        :return:
        :rtype:
        """

        mock_get_fleet.return_value = Fleet(fleet_id=self.fleet_id)

        self.client.force_login(self.user_with_basic_acces_perms)
        url = reverse("fleetfinder:fleet_details", args=[self.fleet_id])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    def test_fleet_redirects_to_dashboard_if_not_found(self):
        """
        Test that the fleet_details view redirects to the dashboard if the fleet does not exist.

        :return:
        :rtype:
        """

        self.client.force_login(self.user_with_manage_perms)

        response = self.client.get(reverse("fleetfinder:fleet_details", args=[123]))

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("fleetfinder:dashboard"))


class TestAjaxFleetDetailsView(FleetfinderTestViews):
    """
    Test the ajax_fleet_details view in the Fleet Finder application.
    This view is responsible for returning fleet details in JSON format.
    It should return the fleet members and their ship types, or an empty list if the fleet is empty.
    """

    @patch("fleetfinder.views.get_fleet_composition")
    def test_returns_correct_fleet_details(self, mock_get_fleet_composition):
        """
        Test that the ajax_fleet_details view returns the correct fleet details.

        :param mock_get_fleet_composition:
        :type mock_get_fleet_composition:
        :return:
        :rtype:
        """

        mock_get_fleet_composition.return_value = SimpleNamespace(
            fleet=[{"name": "Pilot1"}, {"name": "Pilot2"}, {"name": "Pilot3"}],
            aggregate={"Frigate": 2, "Cruiser": 1},
        )

        self.client.force_login(user=self.user_with_manage_perms)

        url = reverse("fleetfinder:ajax_fleet_details", args=[self.fleet_id])
        response = self.client.get(url)

        expected_fleet_composition = {
            "fleet_member": [
                {"name": "Pilot1"},
                {"name": "Pilot2"},
                {"name": "Pilot3"},
            ],
            "fleet_composition": [
                {"ship_type_name": "Frigate", "number": 2},
                {"ship_type_name": "Cruiser", "number": 1},
            ],
        }

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(json.loads(response.content), expected_fleet_composition)

    @patch("fleetfinder.views.get_fleet_composition")
    def test_handles_empty_fleet(self, mock_get_fleet_composition):
        """
        Test that the ajax_fleet_details view handles an empty fleet correctly.

        :param mock_get_fleet_composition:
        :type mock_get_fleet_composition:
        :return:
        :rtype:
        """

        mock_get_fleet_composition.return_value = SimpleNamespace(
            fleet=[],
            aggregate={},
        )

        self.client.force_login(user=self.user_with_manage_perms)

        url = reverse("fleetfinder:ajax_fleet_details", args=[self.fleet_id])
        response = self.client.get(url)

        expected_fleet_composition = {"fleet_member": [], "fleet_composition": []}

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(json.loads(response.content), expected_fleet_composition)

    @patch("fleetfinder.views.get_fleet_composition")
    def test_returns_error_when_fleet_does_not_exist(self, mock_get_fleet_composition):
        """
        Test that the ajax_fleet_details view returns an error when the fleet does not exist.

        :param mock_get_fleet_composition:
        :type mock_get_fleet_composition:
        :return:
        :rtype:
        """

        mock_get_fleet_composition.side_effect = Fleet.DoesNotExist

        self.client.force_login(user=self.user_with_manage_perms)

        url = reverse("fleetfinder:ajax_fleet_details", args=[123])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertJSONEqual(
            response.content,
            {"error": "Fleet with ID 123 does not exist."},
        )

    @patch("fleetfinder.views.get_fleet_composition")
    def test_returns_error_when_runtime_error_occurs(self, mock_get_fleet_composition):
        """
        Test that the ajax_fleet_details view returns an error when a runtime error occurs.

        :param mock_get_fleet_composition:
        :type mock_get_fleet_composition:
        :return:
        :rtype:
        """

        mock_get_fleet_composition.side_effect = RuntimeError("Unexpected error")

        self.client.force_login(user=self.user_with_manage_perms)

        url = reverse("fleetfinder:ajax_fleet_details", args=[123])
        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertJSONEqual(
            response.content,
            {"error": "Error retrieving fleet composition: Unexpected error"},
        )
